#define STRICT
#define DIRECTINPUT_VERSION 0x0800
#include <windows.h>
#include <d3dx9.h>
#include <dinput.h>

#include "system.h"
#include "model.h"

#ifdef _DEBUG
	const int	fullscreen	= 0;
#else
	const int	fullscreen	= 1;
#endif
const int	screenWidth	= 1024	;
const int	screenHeight	= 768	;
const float	fieldOfViewY	= D3DX_PI / 4	;
const float	depthOfView	= 400.0f	;

//[of]:Vertex
struct Vertex
{
	float	px	;
	float	py	;
	float	pz	;
			
	float	nx	;
	float	ny	;
	float	nz	;
			
	float	tu	;
	float	tv	;
};

#define VertexFormat (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_TEX1)

//[cf]
//[of]:App
#define numIndices	((N - 1) * (N - 1) * 2 * 3 * 2)
#define numVertices	(N2 + (N - 1) * (N - 1))

class App : public GraphicsDevice
{
private:
	InputManager	inputManager	;
	Texture*	texture	;
	Texture	texture1	;
	Texture	texture2	;
	Light	light	;
	Material	mat	;
	VertexBuffer	vertices	;
	IndexBuffer	indices	;
	float	rotation	;

	void	loadVertices	();
	void	loadIndices	();

protected:
	virtual bool restore	();
	virtual void invalidate	();
	virtual void move	();
	virtual void render	();
	
public:
	int run	();
};

//[of]:loadVertices
void App::loadVertices()
{
	Vertex* v = (Vertex*)vertices.lock();
	
	int i = 0;
	int col, line;
	for(line=0; line<N; line++)
	{
		for(col=0; col<N; col++)
		{
			v[i].px	= s[i].x;
			v[i].py	= s[i].y;
			v[i].pz	= s[i].z;
			v[i].tu	= ((float)col) / (N - 1) * 1.0f;
			v[i].tv	= ((float)line) / (N - 1) * 1.0f;
			i++;
		}
	}

	i = 0;
	for(line=0; line<N; line++)
	{
		for(col=0; col<N; col++)
		{
			int i1	= (col == 0) ? i : i-1;
			int i2	= (col == N-1) ? i : i+1;
			int j1	= (line == 0) ? i : i-N;
			int j2	= (line == N-1) ? i : i+N;
			
			Vector n(	(v[i2].pz-v[i1].pz)*(v[j2].py-v[j1].py)	- (v[j2].pz-v[j1].pz)*(v[i2].py-v[i1].py), 
				((v[i2].px-v[i1].px)*(v[j2].pz-v[j1].pz)	- (v[j2].px-v[j1].px)*(v[i2].pz-v[i1].pz)), 
				- ((v[i2].py-v[i1].py)*(v[j2].px-v[j1].px)	- (v[j2].py-v[j1].py)*(v[i2].px-v[i1].px)));
			n.normalize();
			
			v[i].nx	= n.x;
			v[i].ny	= n.y;
			v[i].nz	= n.z;
			
			i++;
		}
	}

	int j = 0;
	for(line=0; line<N-1; line++)
	{
		for(col=0; col<N-1; col++)
		{
			int i1	= j;
			int i2	= j + 1;
			int i3	= j + N;
			int i4	= j + N + 1;
			
			v[i].px	= (s[i1].x+s[i2].x+s[i3].x+s[i4].x) / 4.0f;
			v[i].py	= (s[i1].y+s[i2].y+s[i3].y+s[i4].y) / 4.0f; 
			v[i].pz	= (s[i1].z+s[i2].z+s[i3].z+s[i4].z) / 4.0f;
			
			Vector n( 	(v[i1].nx + v[i2].nx + v[i3].nx + v[i4].nx),
				(v[i1].ny + v[i2].ny + v[i3].ny + v[i4].ny),
				(v[i1].nz + v[i2].nz + v[i3].nz + v[i4].nz)	);
			n.normalize();
			
			v[i].nx	= n.x;
			v[i].ny	= n.y;
			v[i].nz	= n.z;
			
			v[i].tu	= ((float)col + 0.5f) / (N - 1) * 1.0f;
			v[i].tv	= ((float)line + 0.5f) / (N - 1) * 1.0f;
			i++;
			j++;
		}
		j++;
	}

	vertices.unlock();
}
//[cf]
//[of]:loadIndices
void App::loadIndices()
{
	WORD* buffer = indices.lock();
	int line, col;
	for(line=0; line<N-1; line++)
	{
		for(col=0; col<N-1; col++)
		{
			int base	= line * N + col;
			int base2	= line * (N - 1) + col + N2;
			
			*buffer++ = base; 
			*buffer++ = base + N;
			*buffer++ = base2;
			
			*buffer++ = base + 1; 
			*buffer++ = base2;
			*buffer++ = base + N + 1;
			
			*buffer++ = base;
			*buffer++ = base2;
			*buffer++ = base + 1;
			
			*buffer++ = base + N; 
			*buffer++ = base + N + 1;
			*buffer++ = base2;
		}
	}
	
	indices.unlock();
}
//[cf]

//[of]:restore
bool App::restore()
{
	Device* dev = device();
	
	dev->SetLight	(0, light.handle());
	dev->LightEnable	(0, TRUE);
	dev->SetMaterial	(mat.handle());
	
	dev->ShowCursor(FALSE);
	dev->SetRenderState(D3DRS_LIGHTING,	TRUE );
	dev->SetRenderState(D3DRS_SPECULARENABLE,	FALSE );
	dev->SetRenderState(D3DRS_ZENABLE,	TRUE );
	dev->SetRenderState(D3DRS_DITHERENABLE,	TRUE );
	dev->SetRenderState(D3DRS_AMBIENT,	0x00606060);
	dev->SetRenderState(D3DRS_AMBIENTMATERIALSOURCE,	D3DMCS_COLOR1);
	
	if( !vertices.create(dev, numVertices * sizeof(Vertex), VertexFormat, false) )
		return false;

	if( !indices.create(dev, numIndices * sizeof(WORD)) )
		return false;

	loadIndices();
	
	return true;
}
//[cf]
//[of]:invalidate
void App::invalidate()
{
	vertices.release();
	indices.release();
}
//[cf]
//[of]:move
void App::move()
{
	inputManager.update();
	
	updateModel();
	
	rotation += pi2 / 2000.0f;

	if( inputManager.isKeyDown(DIK_F1) )
		texture = &texture1;
	else if( inputManager.isKeyDown(DIK_F2) )
		texture = &texture2;
}
//[cf]
//[of]:render
void App::render()
{
	loadVertices();

	Device* dev = device();

	// World
	D3DXMATRIX	matWorld;
	D3DXMatrixIdentity(&matWorld);
	D3DXMatrixRotationY(&matWorld, rotation);
	dev->SetTransform(D3DTS_WORLD,	&matWorld);
	
	// View
	D3DXMATRIX	matView;
	D3DXVECTOR3	eye	( 0.0f,	100.0f - 30.0f,	-250.0f	);
	D3DXVECTOR3	lookAt	( 0.0f,	0.0f - 30.0f,	0.0f	);
	D3DXVECTOR3	up	( 0.0f,	1.0f,	0.0f	);
	D3DXMatrixLookAtLH(&matView, &eye, &lookAt, &up);
	dev->SetTransform(D3DTS_VIEW,	&matView);

	// Transform matrices
	D3DXMATRIXA16	matProj;
	D3DXMatrixPerspectiveFovLH(&matProj, fieldOfViewY, (float)::screenWidth / (float)::screenHeight, 0.5f, depthOfView);
	dev->SetTransform(D3DTS_PROJECTION,	&matProj);

	if FAILED(dev->BeginScene())
		return;

	dev->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0);

	// Render the tablecloth
	dev->SetTexture(0, texture->handle);
	dev->SetTextureStageState(0,	D3DTSS_COLOROP,	D3DTOP_MODULATE	);
	dev->SetTextureStageState(0,	D3DTSS_COLORARG1,	D3DTA_TEXTURE	);
	dev->SetTextureStageState(0,	D3DTSS_COLORARG2,	D3DTA_DIFFUSE	);
	dev->SetTextureStageState(0,	D3DTSS_ALPHAOP,	D3DTOP_DISABLE	);
	
	dev->SetSamplerState(0,	D3DSAMP_MAGFILTER,	D3DTEXF_LINEAR);
	dev->SetSamplerState(0,	D3DSAMP_MINFILTER,	D3DTEXF_LINEAR);
	dev->SetSamplerState(0,	D3DSAMP_MIPFILTER,	D3DTEXF_LINEAR);
	
	dev->SetFVF(VertexFormat);
	dev->SetStreamSource(0, vertices.handle, 0, sizeof(Vertex));
	dev->SetIndices(indices.handle);
	dev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, numVertices, 0, numIndices / 3);

	dev->EndScene();
	dev->Present(NULL, NULL, NULL, NULL);
}
//[cf]

//[of]:run
int App::run()
{
	// Initialize screen
	if( !initialize("Tablecloth Simulator", ::screenWidth, ::screenHeight, fullscreen) )
		return 1;
	
	// Initialize keyboard
	if( !inputManager.initialize() )
		return 1;
	
	// Initialize textures
	if( !texture1.load(device(), "red.jpg", 512, 512, D3DFMT_R8G8B8) )
		return 1;
	if( !texture2.load(device(), "carpet.jpg", 512, 512, D3DFMT_R8G8B8) )
		return 1;
	texture = &texture1;
	
	// Initialize light
	light.setDirection(Vector(-1.3f, -1.0f, 1.3f));
	
	// Initialize the model
	initializeModel();
	
	// Initialize view
	rotation = 0.0f;
	
	main();
	return 0;
}
//[cf]
//[cf]

int run()
{
	App app;
	return app.run();
}
