#include <stdlib.h>
#include <math.h>
#include "model.h"

#define SQRT
#define vertex(col, line) ((col) + (line)*n)

struct link
{
	int	a	;
	int	b	;
	float	d	;
	int	t	;
};

typedef struct object	object	;
typedef struct link	link	;
typedef struct object*	System	;

//[c]
//[c]	Parameters
//[c]	

#define MAXN 10000

int	LN	=	0;

#define R  5.0f

float	gravity	= 0.005f / 2	;
float	frotte	= 0.002f	;
float	ressort	= 0.05f	;

object	s[MAXN];
link	l[100000];
object	tmp[MAXN];

float	wind	= 0.0f	;
float	windtar	= 0.0f	;
int	winddelay	= 300	;
int	windflag	= 1	;

// Forces
float	ax[MAXN];
float	ay[MAXN];
float	az[MAXN];

//[of]:process
//[c]
//[c]	Compute acceleration
//[c]
static void process(System target, System initial, System source)
{
	int i;
	float dx, dy, dz, op, rx, ry, rz;
  
	// initialize
	if( target != initial )
	{
		for(i=0; i<N2; i++)
		{
			// copy position and velocity
			target[i].x	= initial[i].x;
			target[i].y	= initial[i].y;
			target[i].z	= initial[i].z;
			target[i].vx	= initial[i].vx;
			target[i].vy	= initial[i].vy;
			target[i].vz	= initial[i].vz;
		}
	}
  
	for(i=0; i<N2; i++)
	{
		// apply viscosity and gravity
		ax[i] = -frotte*source[i].vx + wind / 4.0f;
		ay[i] = -frotte*source[i].vy - gravity;
		az[i] = -frotte*source[i].vz - wind / 2.0f;
	}
  
	for(i=0;i<LN;i++) 
	{
		int a = l[i].a;
		int b = l[i].b;
		
		dx = source[b].x - source[a].x;
		dy = source[b].y - source[a].y;
		dz = source[b].z - source[a].z;
		
		op = SQRT(dx*dx+dy*dy+dz*dz);
		
		if( op!=0 ) 
		{
			// distance
			float r = l[i].d*l[i].d;
			
			// spring constant
			float k = (l[i].t!=2) ? ressort : ressort / 4.0f;
			
			rx = k * (r*dx/op - dx);
			ry = k * (r*dy/op - dy);
			rz = k * (r*dz/op - dz);
			
			ax[a] -= rx;
			ay[a] -= ry;
			az[a] -= rz;
			ax[b] += rx;
			ay[b] += ry;
			az[b] += rz;
		}
	}
  
	// velocity and position
	for(i=0; i<N2; i++)
	{
		if( initial[i].b ) continue;
		
		target[i].vx	+= ax[i]*initial[i].m;
		target[i].vy	+= ay[i]*initial[i].m;
		target[i].vz	+= az[i]*initial[i].m;
		target[i].x	+= target[i].vx;
		target[i].y	+= target[i].vy;
		target[i].z	+= target[i].vz;
		
		if(	target[i].y<0 &&
			target[i].y>-10 &&
			target[i].x*target[i].x+target[i].z*target[i].z<(65*65) )
		{ 
			target[i].y -= target[i].vy; target[i].vy = 0;
			target[i].x -= target[i].vx; target[i].vx /= 2;
			target[i].z -= target[i].vz; target[i].vz /= 2;
		}
	}
}
//[cf]
//[of]:initializeModel
void initializeModel()
{
	int	n	=	N;
	int	col, line;
  
	// create vertices
	for(col=0; col<n; col++)
		for(line=0; line<n; line++)
		{
			int i = col + line*n;
			s[i].x	= col*R - n/2*R	;
			s[i].y	= 0	;
			s[i].z	= line*R - n / 2 * R	;
			s[i].vx	= 0	;
			s[i].vy	= 0	;
			s[i].vz	= 0	;
			s[i].m	 = 1.0f / 4	;
			s[i].b	= 0	;
		}
	
	// Create springs
	int i = 0;
	for(line=0; line<n; line++)
		for(col=0; col<n-1; col++)
		{
			l[i].a	= vertex(col, line)	;
			l[i].b	= vertex(col + 1, line)	;
			l[i].d	= R	;
			l[i].t	= 0	;
			i++;
		}
    
	for(line=0; line<n-1; line++)
		for(col=0; col<n; col++)
		{
			l[i].a	= vertex(col, line)	;
			l[i].b	= vertex(col, line + 1)	;
			l[i].d	= R	;
			l[i].t	= 0	;
			i++;
		}
    
	for(line=0; line<n-1; line++)
		for(col=0; col<n-1; col++)
		{
			l[i].a	= vertex(col, line)	;
			l[i].b	= vertex(col + 1, line + 1)	;
			l[i].d	= R*sqrtf(2.0f)	;
			l[i].t	= 1	;
			i++;
			
			l[i].a	= vertex(col + 1, line)	;
			l[i].b	= vertex(col, line + 1)	;
			l[i].d	= R*sqrtf(2.0f)	;
			l[i].t	= 1	;
			i++;
		}
    
	for(line=0; line<n; line++)
		for(col=0; col<n-2; col++)
		{
			l[i].a	= vertex(col, line)	;
			l[i].b	= vertex(col + 2, line)	;
			l[i].d	= 2 * R	;
			l[i].t	= 2	;
			i++;
		}
	
	for(line=0; line<n-2; line++)
		for(col=0; col<n; col++)
		{
			l[i].a	= vertex(col, line)	;
			l[i].b	= vertex(col, line + 2)	;
			l[i].d	= 2 * R	;
			l[i].t	= 2	;
			i++;
		}
    
	LN	= i;
}
//[cf]
//[of]:updateModel
void updateModel()
{
	if( windflag )
	{
		if( wind>windtar )
		{ 
			wind	-=	0.001f;
		}
		else
		{
			wind	+=	0.001f;
		}
	}
	else
	{
		wind = 0;
	}
	
	winddelay--;
	if( winddelay == 0 )
	{
		winddelay	=	100;
		windtar	=	(float)rand() / 32768 / 200;
	}
	
	for(int jj=0;jj<4;jj++)
	{
		process(tmp, s, s);
		process(s, s, tmp);
	}
}
//[cf]
